#The calculation of these phase-shifts and rotation is derived from the publications below
#Nondiffracting vector beams where the charge and the polarization state vary with propagation distance
#doi.org/10.1364/OL.41.002270
#Arbitrary spin to orbital angular momentum conversion of light
#DOI 10.1126/science.aao5392
from gdsCAD import *
import numpy
import shapely
from shapely.geometry import Polygon
from gdsii.library import Library
from gdsii.elements import *
import csv

#define some required functions
def cart2pol(x, y):
    rho = numpy.sqrt(x**2 + y**2)
    phi = numpy.arctan2(y, x)
    return(rho, phi)

def pol2cart(rho, phi):
    x = rho * numpy.cos(phi)
    y = rho * numpy.sin(phi)
    return(x, y)

def totuple(a):
    try:
        return tuple(totuple(i) for i in a)
    except TypeError:
        return a

def round_2(x, base=2):
    return int(base*numpy.round(float(x)/base))

def round_1(x, base=1):
    return int(base*numpy.round(float(x)/base))

def round_05(x, base=0.5):
    return int(base*numpy.round(float(x)/base))

def round_025(x, base=0.25):
    return int(base*numpy.round(float(x)/base))

def round_0001(x, base=0.001):
    return int(base*numpy.round(float(x)/base))

def round_20(x, base=20):
    return int(base*numpy.round(float(x)/base))

def round_80(x, base=40):
    return int(base*numpy.round(float(x)/base))

def interval_360(x, base=360):
    P = (float(x)/base)
    i, d = divmod(P, 1)
    return round_1(int(float(x) - (i*base) ))

def interval_180(x, base=180):
    P = (float(x)/base)
    i, d = divmod(P, 1)
    return round_1(int(float(x) - (i*base) ))

def interval_1(x, base=1):
    P = (float(x)/base)
    i, d = divmod(P, 1)
    return (int(i*base) )

#Create meta-atom Library below (directly load from 2-column Text-file with , seperator)
library = numpy.loadtxt(open("W-P-T-PillarResults_aniso_H150um_PureSi-15-30.csv", "rb"), delimiter=",", skiprows=1)


for tupl in library:
    #print(tupl)
    file = open("Input Library-Library.txt","a")
    file.write('%d' % (1000*tupl[0]) + ', ' '%d' % (1000*tupl[1]) + ', ' '%d' % (1000*tupl[2]) +  ', ' '%d' % (1000*tupl[3]) + ', ' '%d' % (1000*tupl[4]) + ', ' '%d' % (1000*tupl[5]) + "\n")
    file.close()

N_diameters = len(library)
#del(Tupels)
#print ("Phase Library (Phase-shift, Radius):")
print("Number of Entries in Phase_library:")
print (N_diameters)

#fundamental pattern parameter
nnd = float(40)
size = round_80(1000 * float(input("Which size (diameter in mm)?:" )))
print(size)
periods = int(size/(nnd))
print(periods)

f = 1000*float(input("Which focal length for all Stokes-Points (in mm)?:" ))
df = 1000*float(input("Which transversal off-axis length (from center) of the focal spot for all Stokes-Points (in mm)?:" ))
wl = float(118.8)

i_x = (0)
i_y = (0)
n = int(0)

#Crearte Cells, Lists & containers
LAT1 = []
LAT2 = []
LAT3 = []

Unit_cell = core.Cell('Rectangles')
topcell=core.Cell('TOP')
layout = core.Layout('LIBRARY')
#ellipse_1 = shapes.Ellipse((0,0), 18, 7)
#Unit_cell.add(ellipse_1)

while i_y < (periods/numpy.sqrt(3)):
    while i_x < periods/3:
        resc_y = 0.5*nnd*numpy.sqrt(3)*interval_1((periods/numpy.sqrt(3)))
        lat1 = ((-0.5*size +(i_x*3*nnd)),(-resc_y +(i_y*nnd*numpy.sqrt(3))))
        lat1_2 = ((-0.5*size +((1.5*nnd+i_x*3*nnd))),((-resc_y +(0.5+i_y)*nnd*numpy.sqrt(3))))
        LAT1.append(lat1)
        LAT1.append(lat1_2)
        lat2 = ((-0.5*size +((nnd+i_x*3*nnd))),(-resc_y +(i_y*nnd*numpy.sqrt(3))))
        lat2_2 = ((-0.5*size +((2.5*nnd+i_x*3*nnd))),((-resc_y +(0.5+i_y)*nnd*numpy.sqrt(3))))
        LAT2.append(lat2)
        LAT2.append(lat2_2)
        lat3 = ((-0.5*size +((2*nnd+i_x*3*nnd))),(-resc_y +(i_y*nnd*numpy.sqrt(3))))
        lat3_2 = ((-0.5*size +((0.5*nnd+i_x*3*nnd))),((-resc_y +(0.5+i_y)*nnd*numpy.sqrt(3))))
        LAT3.append(lat3)
        LAT3.append(lat3_2)
        file = open("CombinedLattices_dia2,4_Foc3.2_df0,6_PureSi.txt","a")
        file.write('%d' % lat1[0] + ', ' '%d' % lat1[1] + ', ' '%d' % lat2[0] + ', ' '%d' % lat2[1] + ', ' '%d' % lat3[0] + ', ''%d' % lat3[1] + "\n")
        file.write('%d' % lat1_2[0] + ', ' '%d' % lat1_2[1] + ', ' '%d' % lat2_2[0] + ', ' '%d' % lat2_2[1] + ', ' '%d' % lat3_2[0] + ', ''%d' % lat3_2[1] + "\n")
        file.close()
        i_x = i_x+1
    i_x = 0
    i_y = i_y+1

#Generate lattice points in list of (x,y)-tuples COMs
#for ix in range(-(periods/2), (periods/2)):
#    for iy in range(round_1(-(numpy.sqrt(3)*periods/2)), round_1(numpy.sqrt(3)*periods/2)):
#        lat1 = ( (ix*(a1[0]+a2[0])), (iy*(a1[1]+a2[1])) )
#        lat2 = ( (a2[0]+ix*(a1[0]+a2[0])), (a2[1]+iy*(a1[1]+a2[1])) )
#        lat3 = ( (a1[0]+ix*(a1[0]+a2[0])), (a1[1]+iy*(a1[1]+a2[1])) )
#        LAT1.append(lat1)
#        LAT2.append(lat2)
#        LAT3.append(lat3)
#        lat1_2 = ( ( (0.5+ix)*(a1[0]+a2[0])), ( (0.5+iy)*(a1[1]+a2[1])) )
#        lat2_2 = ( (a2[0]+(0.5+ix)*(a1[0]+a2[0])), (a2[1]+(0.5+iy)*(a1[1]+a2[1])) )
#        lat3_2 = ( (a1[0]+(0.5+ix)*(a1[0]+a2[0])), (a1[1]+(0.5+iy)*(a1[1]+a2[1])) )
#        LAT1.append(lat1_2)
#        LAT2.append(lat2_2)
#        LAT3.append(lat3_2)
#        file = open("Lattices.txt","a")
#        file.write('%d' % lat1[0] + ', ' '%d' % lat1[1] + ', ' '%d' % lat2[0] + ', ' '%d' % lat2[1] + ', ' '%d' % lat3[0] + ', ''%d' % lat3[1] + "\n")
#        file.write('%d' % lat1_2[0] + ', ' '%d' % lat1_2[1] + ', ' '%d' % lat2_2[0] + ', ' '%d' % lat2_2[1] + ', ' '%d' % lat3_2[0] + ', ''%d' % lat3_2[1] + "\n")
#        file.close()


#Define required Phase-profile (x, Phi_x, y, Phi_y) at each lattice point (x,y)
#Phase_req_x & Phase_req_y can encode 2 COMPLETELY DIFFERENT Phase functions onto linear x & y polarizations
#e.g. Phase_req_x generates Axicon with Vortex while Phase_req_y generates a focussing lens
#Angle_scale = (24*numpy.pi)

#MetaLens: Phase_req_x = interval_360(round_225(numpy.rad2deg(-((numpy.pi/wl) * (f_x -  numpy.sqrt( numpy.square(f_x) + numpy.square(COM[0]) + numpy.square(COM[1])) )) )))
#Off-Axis Metalens: Phase_req_x = interval_360(round_225(numpy.rad2deg(-((numpy.pi/wl) * (f_x -  numpy.sqrt( numpy.square(f_x) + numpy.square(COM[0]-xf) + numpy.square(COM[1]-yf)) )) )))
Phase_map1 = []
Phase_map2 = []
Phase_map3 = []
j= 0 + 1j
#Lattice1: X vs. Y pol.
for lat in LAT1:
    polar = cart2pol(lat[0],lat[1])
    angle_1 = 0
    Phase_req_x = (180 - interval_360(round_1(numpy.rad2deg(-(2*numpy.pi/wl)*(-f + numpy.sqrt( numpy.square(f) + numpy.square( lat[0]-(0.5*df) ) + numpy.square( lat[1]+(numpy.sqrt(3)*0.5*df) ) ) ) ) ) ) )
    Phase_req_y = (180 - interval_360(round_1(numpy.rad2deg(-(2*numpy.pi/wl)*(-f + numpy.sqrt( numpy.square(f) + numpy.square(lat[0]+(0.5*df)) + numpy.square(lat[1]-(numpy.sqrt(3)*0.5*df) ) ) ) ) ) ) )
    #continue pick and place algorithm here
    #print(Phase_req_x)
    #print(Phase_req_y)
    delta = 180
    L_opt_x = 0
    L_opt_y = 0
    for tup in library:
        delta_new_x = ( numpy.square(numpy.absolute(tup[2]*(numpy.exp(j*numpy.radians(tup[1]))) - numpy.exp((j*numpy.radians(Phase_req_x) )))) )
        delta_new_y = ( numpy.square(numpy.absolute(tup[5]*(numpy.exp(j*numpy.radians(tup[4]))) - numpy.exp(j*numpy.radians(Phase_req_y)))) )
        #print(numpy.radians(Phase_req_x))
        if (delta_new_x + delta_new_y) < delta:
            delta = (delta_new_x + delta_new_y)
            L_opt_x = tup[0]
            Phase_opt_x = tup[1]
            Trans_opt_x = tup[2]
            L_opt_y = tup[3]
            Phase_opt_y = tup[4]
            Trans_opt_y = tup[5]
    rectangle = shapes.Rectangle( ( ((lat[0]+L_opt_x/2)), (lat[1] + (L_opt_y)/2) ) , ( ((lat[0]-L_opt_x/2)), (lat[1] - (L_opt_y)/2) ) )
    topcell.add(rectangle)
    file = open("LatXY_2,4_Foc3.2_df0,6_PureSi.txt","a")
    file.write('%d' % lat[0] + ', ' '%d' % L_opt_x + ', ' '%d' % Phase_req_x + ', ' '%d' % Phase_opt_x + ', ' '%d' % lat[1] + ', ''%d' % L_opt_y + ', ''%d' % Phase_req_y + ', ''%d' % Phase_opt_y + ', ''%d' % angle_1 + "\n")
    file.close()

#Lattice2: +V vs -V pol.

for lat in LAT2:
    polar = cart2pol(lat[0],lat[1])
    angle_2 = 45
    Phase_req_x = (180 - interval_360(round_1(numpy.rad2deg(-(2*numpy.pi/wl)*(-f + numpy.sqrt( numpy.square(f) + numpy.square(lat[0]-(0.5*df)) + numpy.square(lat[1]-(numpy.sqrt(3)*0.5*df) ) ) ) ) ) ) )
    Phase_req_y = (180 - interval_360(round_1(numpy.rad2deg(-(2*numpy.pi/wl)*(-f + numpy.sqrt( numpy.square(f) + numpy.square(lat[0]+(0.5*df)) + numpy.square(lat[1]+(numpy.sqrt(3)*0.5*df)) ) ) ) ) ) )
    #continue pick and place algorithm here
    #print(Phase_req_x)
    #print(Phase_req_y)
    delta = 180
    L_opt_x = 0
    L_opt_y = 0
    for tup in library:
        delta_new_x = ( numpy.square(numpy.absolute(tup[2]*(numpy.exp(j*numpy.radians(tup[1]))) - numpy.exp((j*numpy.radians(Phase_req_x) )))) )
        delta_new_y = ( numpy.square(numpy.absolute(tup[5]*(numpy.exp(j*numpy.radians(tup[4]))) - numpy.exp(j*numpy.radians(Phase_req_y)))) )
        #print(numpy.radians(Phase_req_x))
        if (delta_new_x + delta_new_y) < delta:
            delta = (delta_new_x + delta_new_y)
            L_opt_x = tup[0]
            Phase_opt_x = tup[1]
            Trans_opt_x = tup[2]
            L_opt_y = tup[3]
            Phase_opt_y = tup[4]
            Trans_opt_y = tup[5]
    rectangle = shapes.Rectangle( ( ((lat[0]+L_opt_x/2)), (lat[1] + (L_opt_y)/2) ) , ( ((lat[0]-L_opt_x/2)), (lat[1] - (L_opt_y)/2) ) )
    struc = rectangle.rotate(angle_2, center=(lat[0],lat[1]))
    topcell.add(struc)
    file = open("Lat+-45_dia2,4_Foc3.2_df0,6_PureSi.txt","a")
    file.write('%d' % lat[0] + ', ' '%d' % L_opt_x + ', ' '%d' % Phase_req_x + ', ' '%d' % Phase_opt_x + ', ' '%d' % lat[1] + ', ''%d' % L_opt_y + ', ''%d' % Phase_req_y + ', ''%d' % Phase_opt_y + ', ''%d' % angle_2 + "\n")
    file.close()

#Lattice3: RCP vs. LCP pol.


for lat in LAT3:
    polar = cart2pol(lat[0],lat[1])
    Lens_R = ((numpy.rad2deg(-(2*numpy.pi/wl)*(-f + numpy.sqrt( numpy.square(f) + numpy.square(lat[0]-(df)) + numpy.square(lat[1]) ) ) ) ) )
    Lens_L = ((numpy.rad2deg(-(2*numpy.pi/wl)*(-f + numpy.sqrt( numpy.square(f) + numpy.square(lat[0]+(df)) + numpy.square(lat[1]) ) ) ) ) )
    Phase_req_x = (180 - interval_360(round_1(0.5* (Lens_R+Lens_L) ) ))
    Phase_req_y = (180 - interval_360(round_1(-180 + 0.5* (Lens_R+Lens_L) ) ))
    angle_3 = round_05((((0.25*(Lens_R-Lens_L)) )))
    #continue pick and place algorithm here
    #print(Phase_req_x)
    #print(Phase_req_y)
    delta = 180
    L_opt_x = 0
    L_opt_y = 0
    for tup in library:
        delta_new_x = ( numpy.square(numpy.absolute(tup[2]*(numpy.exp(j*numpy.radians(tup[1]))) - numpy.exp((j*numpy.radians(Phase_req_x) )))) )
        delta_new_y = ( numpy.square(numpy.absolute(tup[5]*(numpy.exp(j*numpy.radians(tup[4]))) - numpy.exp(j*numpy.radians(Phase_req_y)))) )
        #print(numpy.radians(Phase_req_x))
        if (delta_new_x + delta_new_y) < delta:
            delta = (delta_new_x + delta_new_y)
            L_opt_x = tup[0]
            Phase_opt_x = tup[1]
            Trans_opt_x = tup[2]
            L_opt_y = tup[3]
            Phase_opt_y = tup[4]
            Trans_opt_y = tup[5]
    rectangle = shapes.Rectangle( ( ((lat[0]+L_opt_x/2)), (lat[1] + (L_opt_y)/2) ) , ( ((lat[0]-L_opt_x/2)), (lat[1] - (L_opt_y)/2) ) )
    struc = rectangle.rotate(angle_3, center=(lat[0],lat[1]))
    topcell.add(struc)
    file = open("LatRL_dia2,4_Foc3.2_df0,6_PureSi.txt","a")
    file.write('%d' % lat[0] + ', ' '%d' % L_opt_x + ', ' '%d' % Phase_req_x + ', ' '%d' % Phase_opt_x + ', ' '%d' % lat[1] + ', ''%d' % L_opt_y + ', ''%d' % Phase_req_y + ', ''%d' % Phase_opt_y + ', ''%d' % angle_3 + "\n")
    file.close()


#Create Ellipses from (L_x,L_y) tupels and add to topcell
#for struc in Phase_map:
#    rectangle = shapes.Rectangle( ( ((struc[0]+struc[1]/2)), (struc[2] + (struc[3])/2) ) , ( ((struc[0]-struc[1]/2)), (struc[2] - (struc[3])/2) ) )
#    topcell.add(rectangle)
#ADDITIONAL ROTATION FOR INCLUSION OF GEOMETRIC PHASE BELOW!!!!
#for COM in COMs:
#    polar = cart2pol(COM[0],COM[1])
#    struc = core.CellReference(Unit_cell, origin=(COM[0],COM[1]), rotation=(round_2(0.5*numpy.rad2deg(-Angle_scale + l*polar[1] + numpy.pi*polar[0]/axicon_P))), magnification=None, x_reflection=False)
#    topcell.add(struc)
#    del(struc)

#add everything to layout and save it
layout.add(topcell)
layout.save('Hexa-MP_dia2,4_Foc3.2_df0,6_PureSi.gds')
#layout.show()
